/*$Id: ezmlm-manage.c,v 1.56.1.4 1999/08/08 14:33:16 lindberg Exp $*/
/*$Name: ezmlm-idx-0316 $*/

#include <sys/types.h>
#include <sys/stat.h>
#include "error.h"
#include "stralloc.h"
#include "str.h"
#include "env.h"
#include "sig.h"
#include "slurp.h"
#include "getconf.h"
#include "strerr.h"
#include "byte.h"
#include "getln.h"
#include "case.h"
#include "qmail.h"
#include "substdio.h"
#include "readwrite.h"
#include "seek.h"
#include "quote.h"
#include "datetime.h"
#include "now.h"
#include "date822fmt.h"
#include "fmt.h"
#include "subscribe.h"
#include "cookie.h"
#include "sgetopt.h"
#include "copy.h"
#include "errtxt.h"
#include "idx.h"

#define FATAL "ezmlm-manage: fatal: "
#define INFO "ezmlm-manage: info: "

int flaglist = 0;	/* default: do not reply to -list */
int flagget = 1;	/* default: service -get requests */
int flagsubconf = 1;	/* default: require user-confirm for subscribe */
int flagunsubconf = 1;	/* default: require user-confirm for unsubscribe */
int flagunsubismod = 0;	/* default: do not require moderator approval to */
			/* unsubscribe from moderated list */
int flagedit = 0;	/* default: text file edit not allowed */
char flagcd = '\0';	/* default: do not use _Q_uoted printable or _B_ase64 */
char encin = '\0';	/* encoding of incoming message */
int flagdig = 0;	/* request is not for digest list */
#define FLD_DIGEST 1
#define FLD_ALLOW 2
#define FLD_DENY 3

int act = AC_NONE;	/* desired action */
char *dir;
char *workdir;
char *sender;

void die_usage() {
  strerr_die1x(100,"ezmlm-manage: usage: ezmlm-manage "
		"[-cCdDeElLmMsSuUvV] dir"); }

void die_nomem() { strerr_die2x(111,FATAL,ERR_NOMEM); }

void die_badaddr()
{
  strerr_die2x(100,FATAL,ERR_BAD_ADDRESS);
}

void die_cookie()
{
  strerr_die2x(100,FATAL,ERR_MOD_COOKIE);
}

stralloc inhost = {0};
stralloc outhost = {0};
stralloc inlocal = {0};
stralloc outlocal = {0};
stralloc key = {0};
stralloc mailinglist = {0};
stralloc mydtline = {0};
stralloc target = {0};
stralloc confirm = {0};
stralloc line = {0};
stralloc qline = {0};
stralloc quoted = {0};
stralloc moddir = {0};
stralloc ddir = {0};
stralloc modsub = {0};
stralloc remote = {0};
stralloc from = {0};
stralloc to = {0};
stralloc text = {0};
stralloc fnedit = {0};
stralloc fneditn = {0};
stralloc charset = {0};

datetime_sec when;
struct datetime dt;
int match;
unsigned int max;

char strnum[FMT_ULONG];
char date[DATE822FMT];
char hash[COOKIE];
char boundary[COOKIE];
datetime_sec hashdate;

int hashok(action,ac)
char *action;
char *ac;
{
  char *x;
  datetime_sec u;

  x = action + 4;
  x += scan_ulong(x,&u);
  hashdate = u;
  if (hashdate > when) return 0;
  if (hashdate < when - 1000000) return 0;

  u = hashdate;
  strnum[fmt_ulong(strnum,(unsigned long) u)] = 0;
  cookie(hash,key.s,key.len - flagdig,strnum,target.s,ac+1);

  if (*x == '.') ++x;
  if (str_len(x) != COOKIE) return 0;
  return byte_equal(hash,COOKIE,x);
}

struct qmail qq;
int qqwrite(fd,buf,len) int fd; char *buf; unsigned int len;
{
  qmail_put(&qq,buf,len);
  return len;
}
char qqbuf[1];
substdio ssqq = SUBSTDIO_FDBUF(qqwrite,-1,qqbuf,sizeof(qqbuf));

char inbuf[1024];
substdio ssin = SUBSTDIO_FDBUF(read,0,inbuf,sizeof(inbuf));
substdio ssin2 = SUBSTDIO_FDBUF(read,0,inbuf,sizeof(inbuf));

substdio sstext;
char textbuf[1024];

int code_qput(s,n)
char *s;
unsigned int n;
{
    if (!flagcd)
      qmail_put(&qq,s,n);
    else {
      if (flagcd == 'B')
        encodeB(s,n,&qline,0,FATAL);
      else
        encodeQ(s,n,&qline,FATAL);
      qmail_put(&qq,qline.s,qline.len);
    }
    return 0;		/* always succeeds */
}

void transferenc()
{
	if (flagcd) {
          qmail_puts(&qq,"\n--");
          qmail_put(&qq,boundary,COOKIE);
          qmail_puts(&qq,"\nContent-Type: text/plain; charset=");
          qmail_puts(&qq,charset.s);
	  qmail_puts(&qq,"\nContent-Transfer-Encoding: ");
          if (flagcd == 'Q')
            qmail_puts(&qq,"quoted-printable\n\n");
          else
	    qmail_puts(&qq,"base64\n\n");
        } else
          qmail_puts(&qq,"\n");
}

void msg_headers()
		/* Writes all the headers up to but not including subject */
{
  int flaggoodfield;
  unsigned int pos;

  qmail_puts(&qq,"Mailing-List: ");
  qmail_put(&qq,mailinglist.s,mailinglist.len);
  qmail_puts(&qq,"\nDate: ");
  datetime_tai(&dt,when);
  qmail_put(&qq,date,date822fmt(date,&dt));
  qmail_puts(&qq,"Message-ID: <");
  if (!stralloc_copyb(&line,strnum,fmt_ulong(strnum,(unsigned long) when)))
     die_nomem();
  if (!stralloc_append(&line,".")) die_nomem();
  if (!stralloc_catb(&line,strnum,
		fmt_ulong(strnum,(unsigned long) getpid()))) die_nomem();
  if (!stralloc_cats(&line,".ezmlm@")) die_nomem();
  if (!stralloc_cat(&line,&outhost)) die_nomem();
  if (!stralloc_0(&line)) die_nomem();
  qmail_puts(&qq,line.s);
		/* "unique" MIME boundary as hash of messageid */
  cookie(boundary,"",0,"",line.s,"");
  qmail_puts(&qq,">\nFrom: ");
  if (!quote(&quoted,&outlocal)) die_nomem();
  qmail_put(&qq,quoted.s,quoted.len);
  if (act == AC_HELP)		/* differnt "From:" for help to break auto- */
    qmail_puts(&qq,"-return-@");	/* responder loops */
  else
    qmail_puts(&qq,"-help@");
  qmail_put(&qq,outhost.s,outhost.len);
  qmail_puts(&qq,"\nTo: ");
  if (!quote2(&quoted,target.s)) die_nomem();
  qmail_put(&qq,quoted.s,quoted.len);
  qmail_puts(&qq,"\n");
  if (!stralloc_copys(&mydtline,"Delivered-To: responder for ")) die_nomem();
  if (!stralloc_catb(&mydtline,outlocal.s,outlocal.len)) die_nomem();
  if (!stralloc_cats(&mydtline,"@")) die_nomem();
  if (!stralloc_catb(&mydtline,outhost.s,outhost.len)) die_nomem();
  if (!stralloc_cats(&mydtline,"\n")) die_nomem();
  qmail_put(&qq,mydtline.s,mydtline.len);

  flaggoodfield = 0;
  for (;;) {
    if (getln(&ssin,&line,&match,'\n') == -1)
      strerr_die2sys(111,FATAL,ERR_READ_INPUT);
    if (!match) break;
    if (line.len == 1) break;
    if ((line.s[0] != ' ') && (line.s[0] != '\t')) {
      flaggoodfield = 0;
      if (case_startb(line.s,line.len,"mailing-list:"))
        strerr_die2x(100,FATAL,ERR_MAILING_LIST);
      if (line.len == mydtline.len)
	if (byte_equal(line.s,line.len,mydtline.s))
          strerr_die2x(100,FATAL,ERR_LOOPING);
      if (case_startb(line.s,line.len,"delivered-to:"))
        flaggoodfield = 1;
      else if (case_startb(line.s,line.len,"received:"))
        flaggoodfield = 1;
      else if (case_startb(line.s,line.len,"content-transfer-encoding:")) {
        pos = 26;
        while (line.s[pos] == ' ' || line.s[pos] == '\t') ++pos;
        if (case_startb(line.s+pos,line.len-pos,"base64"))
          encin = 'B';
        else if (case_startb(line.s+pos,line.len-pos,"quoted-printable"))
          encin = 'Q';
      }
    }
    if (flaggoodfield)
      qmail_put(&qq,line.s,line.len);
  }
  qmail_puts(&qq,"MIME-Version: 1.0\n");
  if (flagcd) {
    qmail_puts(&qq,"Content-Type: multipart/mixed; charset=");
    qmail_puts(&qq,charset.s);
    qmail_puts(&qq,";\n\tboundary=");
    qmail_put(&qq,boundary,COOKIE);
  } else {
    qmail_puts(&qq,"Content-type: text/plain; charset=");
    qmail_puts(&qq,charset.s);
  }
  qmail_puts(&qq,"\n");
}

void geton(action)
char *action;
{
  switch(subscribe(workdir,target.s,1,FATAL)) {
    case 1:
            log(workdir,"+",target.s);
            qmail_puts(&qq,TXT_WELCOME);
            if (!quote(&quoted,&outlocal)) die_nomem();
            qmail_put(&qq,quoted.s,quoted.len);
            qmail_puts(&qq,"@");
            qmail_put(&qq,outhost.s,outhost.len);
            qmail_puts(&qq,"\n");
            transferenc();
            copy(&qq,"text/top",flagcd,FATAL);
            copy(&qq,"text/sub-ok",flagcd,FATAL);
            break;
    default: 
            if (str_start(action,ACTION_TC))
              strerr_die2x(0,INFO,ERR_SUB_NOP);
            qmail_puts(&qq,TXT_EZMLM_RESPONSE);
            transferenc();
            copy(&qq,"text/top",flagcd,FATAL);
            copy(&qq,"text/sub-nop",flagcd,FATAL);
            break;
  }
  if (flagdig == FLD_DENY || flagdig == FLD_ALLOW)
    strerr_die3x(0,INFO,ERR_EXTRA_SUB,target.s);
}

void getoff(action)
char *action;
{
  switch(subscribe(workdir,target.s,0,FATAL)) {
    case 1:
            log(workdir,"-",target.s);
            qmail_puts(&qq,TXT_GOODBYE);
            if (!quote(&quoted,&outlocal)) die_nomem();
            qmail_put(&qq,quoted.s,quoted.len);
            qmail_puts(&qq,"@");
            qmail_put(&qq,outhost.s,outhost.len);
            qmail_puts(&qq,"\n\n");
            transferenc();
            copy(&qq,"text/top",flagcd,FATAL);
            copy(&qq,"text/unsub-ok",flagcd,FATAL);
            break;
    default:
            qmail_puts(&qq,TXT_EZMLM_RESPONSE);
            if (str_start(action,ACTION_VC))
              strerr_die2x(0,INFO,ERR_UNSUB_NOP);
            transferenc();
            copy(&qq,"text/top",flagcd,FATAL);
            copy(&qq,"text/unsub-nop",flagcd,FATAL);
            break;
  }
  if (flagdig == FLD_DENY || flagdig == FLD_ALLOW)
    strerr_die3x(0,INFO,ERR_EXTRA_UNSUB,target.s);
}

void doconfirm(act)
/* This should only be called with valid act for sub/unsub confirms. If act */
/* is not ACTION_SC or ACTION_TC, it is assumed to be an unsubscribe conf.*/
char *act;	/* first letter of desired confirm request only as STRING! */
{
  unsigned int i;

  strnum[fmt_ulong(strnum,(unsigned long) when)] = 0;
  cookie(hash,key.s,key.len-flagdig,strnum,target.s,act+1);
  if (!stralloc_copy(&confirm,&outlocal)) die_nomem();
  if (!stralloc_append(&confirm,"-")) die_nomem();
  if (!stralloc_catb(&confirm,act+1,1)) die_nomem();
  if (!stralloc_cats(&confirm,"c.")) die_nomem();
  if (!stralloc_cats(&confirm,strnum)) die_nomem();
  if (!stralloc_append(&confirm,".")) die_nomem();
  if (!stralloc_catb(&confirm,hash,COOKIE)) die_nomem();
  if (!stralloc_append(&confirm,"-")) die_nomem();
  i = str_rchr(target.s,'@');
  if (!stralloc_catb(&confirm,target.s,i)) die_nomem();
  if (target.s[i]) {
    if (!stralloc_append(&confirm,"=")) die_nomem();
    if (!stralloc_cats(&confirm,target.s + i + 1)) die_nomem();
  }
  if (!stralloc_append(&confirm,"@")) die_nomem();
  if (!stralloc_cat(&confirm,&outhost)) die_nomem();
  if (!stralloc_0(&confirm)) die_nomem();
  set_cpconfirm(confirm.s);		/* for copy */

  qmail_puts(&qq,"Reply-To: ");
  if (!quote2(&quoted,confirm.s)) die_nomem();
  qmail_put(&qq,quoted.s,quoted.len);
  qmail_puts(&qq,"\n");
  if (!stralloc_0(&confirm)) die_nomem();

  qmail_puts(&qq,"Subject: ");
  if (act[1] == ACTION_SC[1] || act[1] == ACTION_UC[1])
    qmail_puts(&qq,TXT_USRCONFIRM);
  else
    qmail_puts(&qq,TXT_MODCONFIRM);
  if (act[1] == ACTION_SC[1] || act[1] == ACTION_TC[1])
    qmail_puts(&qq,TXT_SUBSCRIBE_TO);
  else
    qmail_puts(&qq,TXT_UNSUBSCRIBE_FROM);
  if (!quote(&quoted,&outlocal)) die_nomem();
  qmail_put(&qq,quoted.s,quoted.len);
  qmail_puts(&qq,"@");
  qmail_put(&qq,outhost.s,outhost.len);
  qmail_puts(&qq,"\n");
  transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
}

void sendtomods()
{
  tosubs(moddir.s,"",&qq,FATAL);
}

void copybottom()
{
  copy(&qq,"text/bottom",flagcd,FATAL);
  if (flagcd) {
    if (flagcd == 'B') {
      encodeB("",0,&line,2,FATAL);	/* flush */
      qmail_put(&qq,line.s,line.len);
    }
    qmail_puts(&qq,"\n--");
    qmail_put(&qq,boundary,COOKIE);
    qmail_puts(&qq,"\nContent-Type: message/rfc822");
    qmail_puts(&qq,"\nContent-Disposition: inline; filename=request.msg\n\n");
  }
  qmail_puts(&qq,"Return-Path: <");
  if (!quote2(&quoted,sender)) die_nomem();
  qmail_put(&qq,quoted.s,quoted.len);
  qmail_puts(&qq,">\n");
  if (seek_begin(0) == -1)
    strerr_die2sys(111,FATAL,ERR_SEEK_INPUT);
  if (substdio_copy(&ssqq,&ssin2) != 0)
    strerr_die2sys(111,FATAL,ERR_READ_INPUT);
  if (flagcd) {
    qmail_puts(&qq,"\n--");
    qmail_put(&qq,boundary,COOKIE);
    qmail_puts(&qq,"--\n");
  }
  qmail_from(&qq,from.s);
}

void main(argc,argv)
int argc;
char **argv;
{
  char *host;
  char *local;
  char *action;
  char *x, *y;
  char *fname;
  char *pmod;
  char *cp,*cpfirst,*cplast,*cpnext,*cpafter;
  int flagmod;
  int flagremote;
  int flagpublic;
  int opt;
  unsigned int i;
  unsigned int len;
  int fd,fdlock;
  int flagdone;
  register char ch;

  (void) umask(022);
  sig_pipeignore();
  when = now();

  while ((opt = getopt(argc,argv,"cCdDeElLmMsSuUvV")) != opteof)
    switch(opt) {
      case 'c': flagget = 1; break;
      case 'C': flagget = 0; break;
      case 'd':
      case 'e': flagedit = 1; break;
      case 'D':
      case 'E': flagedit = 0; break;
      case 'l': flaglist = 1; break;
      case 'L': flaglist = 0; break;
      case 'm': flagunsubismod = 1; break;
      case 'M': flagunsubismod = 0; break;
      case 's': flagsubconf = 1; break;
      case 'S': flagsubconf = 0; break;
      case 'u': flagunsubconf = 1; break;
      case 'U': flagunsubconf = 0; break;
      case 'v':
      case 'V': strerr_die2x(100,FATAL,"ezmlm-manage version: ezmlm-0.53+$Name: ezmlm-idx-0316 $\n");
      default:
	die_usage();
    }

  dir = argv[optind];
  if (!dir) die_usage();

  sender = env_get("SENDER");
  if (!sender) strerr_die2x(100,FATAL,ERR_NOSENDER);
  local = env_get("LOCAL");
  if (!local) strerr_die2x(100,FATAL,ERR_NOLOCAL);
  host = env_get("HOST");
  if (!host) strerr_die2x(100,FATAL,ERR_NOHOST);

  if (!*sender)
    strerr_die2x(100,FATAL,ERR_BOUNCE);
  if (!sender[str_chr(sender,'@')])
    strerr_die2x(100,FATAL,ERR_ANONYMOUS);
  if (str_equal(sender,"#@[]"))
    strerr_die2x(100,FATAL,ERR_BOUNCE);

  if (chdir(dir) == -1)
    strerr_die4sys(111,FATAL,ERR_SWITCH,dir,": ");

  switch(slurp("key",&key,32)) {
    case -1:
      strerr_die4sys(111,FATAL,ERR_READ,dir,"/key: ");
    case 0:
      strerr_die4x(100,FATAL,dir,"/key",ERR_NOEXIST);
  }
  getconf_line(&mailinglist,"mailinglist",1,FATAL,dir);
  getconf_line(&inhost,"inhost",1,FATAL,dir);
  getconf_line(&inlocal,"inlocal",1,FATAL,dir);
  getconf_line(&outhost,"outhost",1,FATAL,dir);
  getconf_line(&outlocal,"outlocal",1,FATAL,dir);
  set_cpouthost(&outhost);
  if (getconf_line(&charset,"charset",0,FATAL,dir)) {
    if (charset.len >= 2 && charset.s[charset.len - 2] == ':') {
      if (charset.s[charset.len - 1] == 'B' ||
		charset.s[charset.len - 1] == 'Q') {
        flagcd = charset.s[charset.len - 1];
        charset.s[charset.len - 2] = '\0';
      }
    }
  } else
    if (!stralloc_copys(&charset,TXT_DEF_CHARSET)) die_nomem();
  if (!stralloc_0(&charset)) die_nomem();

  if (inhost.len != str_len(host)) die_badaddr();
  if (case_diffb(inhost.s,inhost.len,host)) die_badaddr();
  if (inlocal.len > str_len(local)) die_badaddr();
  if (case_diffb(inlocal.s,inlocal.len,local)) die_badaddr();

  action = local + inlocal.len;

  if (!stralloc_copys(&ddir,dir)) die_nomem();

  if (case_starts(action,ACTION_DIGEST)) {		/* -digest */
    action += str_len(ACTION_DIGEST);
    if (!stralloc_cats(&outlocal,ACTION_DIGEST)) die_nomem();
    if (!stralloc_cats(&ddir,"/digest")) die_nomem();
    flagdig = FLD_DIGEST;
  } else if (case_starts(action,ACTION_ALLOW)) {	/* -allow */
    action += str_len(ACTION_ALLOW);
    if (!stralloc_cats(&outlocal,ACTION_ALLOW)) die_nomem();
    if (!stralloc_cats(&ddir,"/extra")) die_nomem();
    flagdig = FLD_ALLOW;
  } else if (case_starts(action,ACTION_DENY)) {		/* -deny */
    action += str_len(ACTION_DENY);
    if (!stralloc_cats(&outlocal,ACTION_DENY)) die_nomem();
    if (!stralloc_cats(&ddir,"/blacklist")) die_nomem();
    flagdig = FLD_DENY;
  }
  if (!stralloc_0(&ddir)) die_nomem();
  workdir = ddir.s;
  set_cpoutlocal(&outlocal);

  if (!stralloc_copys(&target,sender)) die_nomem();
  if (action[0]) {
    i = 1 + str_chr(action + 1,'-');
    if (action[i]) {
      action[i] = 0;
      if (!stralloc_copys(&target,action + i + 1)) die_nomem();
      i = byte_rchr(target.s,target.len,'=');
      if (i < target.len)
	target.s[i] = '@';
    }
  }
  if (!stralloc_0(&target)) die_nomem();
  set_cptarget(target.s);	/* for copy() */

  flagmod = getconf_line(&modsub,"modsub",0,FATAL,dir);
  flagremote = getconf_line(&remote,"remote",0,FATAL,dir);

  if (case_equals(action,ACTION_LIST) ||
		case_equals(action,ALT_LIST))
    act = AC_LIST;
  else if (case_starts(action,ACTION_GET) ||
		case_starts(action,ALT_GET))
    act = AC_GET;
  else if (case_equals(action,ACTION_HELP) ||
		case_equals(action,ALT_HELP))
    act = AC_HELP;
  else if (case_starts(action,ACTION_EDIT) ||
		case_starts(action,ALT_EDIT))
    act = AC_EDIT;
  else if (case_equals(action,ACTION_LOG) ||
		case_equals(action,ALT_LOG))
    act = AC_LOG;

			/* NOTE: act is needed in msg_headers(). */
			/* Yes, this needs to be cleaned up! */

  if (flagmod || flagremote) {
    if (modsub.len && modsub.s[0] == '/') {
      if (!stralloc_copy(&moddir,&modsub)) die_nomem();
    } else if (remote.len && remote.s[0] == '/') {
      if (!stralloc_copy(&moddir,&remote)) die_nomem();
    } else {
      if (!stralloc_copys(&moddir,dir)) die_nomem();
      if (!stralloc_cats(&moddir,"/mod")) die_nomem();
    }
    if (!stralloc_0(&moddir)) die_nomem();
		/* for these the reply is 'secret' and goes to sender  */
		/* This means that they can be triggered from a SENDER */
		/* that is not a mod, but never send to a non-mod */
    if (act == AC_NONE || flagdig == FLD_DENY)	/* None of the above */
      pmod = issub(moddir.s,sender,FATAL);
				/* sender = moderator? */
    else
      pmod = issub(moddir.s,target.s,FATAL);
				/* target = moderator? */
   } else
     pmod = 0;			/* always 0 for non-mod/remote lists */
				/* if DIR/public is missing, we still respond*/
				/* to requests from moderators for remote    */
				/* admin and modsub lists. Since pmod   */
				/* is false for all non-mod lists, only it   */
				/* needs to be tested. */
  if ((flagpublic = slurp("public",&line,1)) == -1)
      strerr_die4sys(111,FATAL,ERR_READ,dir,"/public: ");
  if (!flagpublic && !(pmod && flagremote) &&
                !case_equals(action,ACTION_HELP))
      strerr_die2x(100,FATAL,ERR_NOT_PUBLIC);

  if (flagdig == FLD_DENY)
    if (!pmod || !flagremote)			/* only mods can do */
      strerr_die1x(100,ERR_NOT_ALLOWED);

  if (!stralloc_copy(&from,&outlocal)) die_nomem();
  if (!stralloc_cats(&from,"-return-@")) die_nomem();
  if (!stralloc_cat(&from,&outhost)) die_nomem();
  if (!stralloc_0(&from)) die_nomem();

  if (qmail_open(&qq) == -1)
    strerr_die2sys(111,FATAL,ERR_QMAIL_QUEUE);
  msg_headers();

  if (case_equals(action,ACTION_SUBSCRIBE)
		|| case_equals(action,ALT_SUBSCRIBE)) {
    if (pmod && flagremote) {
      doconfirm(ACTION_TC);
      copy(&qq,"text/mod-sub-confirm",flagcd,FATAL);
      copybottom();
      qmail_to(&qq,pmod);
    } else if (flagsubconf) {
      doconfirm(ACTION_SC);
      copy(&qq,"text/sub-confirm",flagcd,FATAL);
      copybottom();
      qmail_to(&qq,target.s);
    } else if (flagmod || (flagremote && str_diff(target.s,sender))) {
      doconfirm(ACTION_TC);	/* remote part to be nice: not more secure */
      copy(&qq,"text/mod-sub-confirm",flagcd,FATAL);
      copybottom();
      sendtomods();
    } else {
      geton(action);
      copybottom();
      qmail_to(&qq,target.s);
    }

  } else if (str_start(action,ACTION_SC)) {
    if (hashok(action,ACTION_SC)) {
      if (flagmod && !(pmod && str_equal(sender,target.s))) {
        doconfirm(ACTION_TC);
        copy(&qq,"text/mod-sub-confirm",flagcd,FATAL);
        copybottom();
        sendtomods();
      } else {
        geton(action);
        copybottom();
        qmail_to(&qq,target.s);
      }
    } else {
      doconfirm(ACTION_SC);
      copy(&qq,"text/sub-bad",flagcd,FATAL);
      copybottom();
      qmail_to(&qq,target.s);
    }

  } else if (str_start(action,ACTION_TC)) {
    if (hashok(action,ACTION_TC)) {
      geton(action);
      copy(&qq,"text/mod-sub",flagcd,FATAL);
      copy(&qq,"text/bottom",flagcd,FATAL);
      code_qput(TXT_SUPPRESSED,str_len(TXT_SUPPRESSED));
      if (flagcd) {			/* end boundary */
	qmail_puts(&qq,"\n--");
	qmail_put(&qq,boundary,COOKIE);
	qmail_puts(&qq,"--\n");
      }
      if (flagcd == 'B') {
        encodeB("",0,&line,2,FATAL);	/* flush */
        qmail_put(&qq,line.s,line.len);
      }
      qmail_from(&qq,from.s);
      qmail_to(&qq,target.s);
    } else {
      if (!pmod || !flagremote)	/* else anyone can get a good -tc. */
        die_cookie();
      doconfirm(ACTION_TC);
      copy(&qq,"text/sub-bad",flagcd,FATAL);
      copybottom();
      qmail_to(&qq,pmod);
    }

  } else if (case_equals(action,ACTION_UNSUBSCRIBE) ||
             case_equals(action,ALT_UNSUBSCRIBE)) {

    if (flagunsubconf) {
      if (pmod && flagremote) {
        doconfirm(ACTION_VC);
        copy(&qq,"text/mod-unsub-confirm",flagcd,FATAL);
        copybottom();
        qmail_to(&qq,pmod);
      } else {
        doconfirm(ACTION_UC);
        copy(&qq,"text/unsub-confirm",flagcd,FATAL);
        copybottom();
        qmail_to(&qq,target.s);
      }
    } else if ((flagmod || flagremote) && str_diff(target.s,sender)) {
      				/* to make some sense of remote with -U */
      doconfirm(ACTION_VC);		/* -vc. here is ok. It can only unsub */
      copy(&qq,"text/unsub-confirm",flagcd,FATAL);
					/* and is only sent to mod or target*/
      copybottom();			/* want -vc to suppress trigger on */
      if (pmod && flagremote) {		/* reply, which otherwise would */
        qmail_to(&qq,pmod);		/* reveal moderator address */
      } else {
        qmail_to(&qq,target.s);
        sendtomods();
      }
    } else {
      getoff(action);
      copybottom();
      qmail_to(&qq,target.s);
    }

  } else if (str_start(action,ACTION_UC)) {
    if (hashok(action,ACTION_UC)) {
	/* unsub is moderated only on moderated list if -m unless the */
	/* target == sender == a moderator */
      if (flagunsubismod &&
		flagmod && !(pmod && str_equal(sender,target.s))) {
        doconfirm(ACTION_VC);
        copy(&qq,"text/mod-unsub-confirm",flagcd,FATAL);
        copybottom();
        sendtomods();
      } else {
        getoff(action);
        copybottom();
        qmail_to(&qq,target.s);
      }
    } else {
      doconfirm(ACTION_UC);
      copy(&qq,"text/unsub-bad",flagcd,FATAL);
      copybottom();
      qmail_to(&qq,target.s);
    }

  } else if (str_start(action,ACTION_VC)) {
    if (hashok(action,ACTION_VC)) {
      getoff(action);
      copy(&qq,"text/mod-sub",flagcd,FATAL);
      if (flagcd == 'B') {
        encodeB("",0,&line,2,FATAL);	/* flush */
        qmail_put(&qq,line.s,line.len);
      }
      qmail_from(&qq,from.s);
      qmail_to(&qq,target.s);

    } else {
      if (!pmod || !flagremote)	/* else anyone can get a good -vc. */
        die_cookie();
      doconfirm(ACTION_VC);
      copy(&qq,"text/unsub-bad",flagcd,FATAL);
      copybottom();
      qmail_to(&qq,pmod);
    }

  } else if (act == AC_LIST) {

    if (!flaglist || (!flagmod && !flagremote))
      strerr_die2x(100,FATAL,ERR_NOT_AVAILABLE);
    if (!pmod)
      strerr_die2x(100,FATAL,ERR_NOT_ALLOWED);
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    (void) code_qput(TXT_LISTMEMBERS,str_len(TXT_LISTMEMBERS));
    i = putsubs(workdir,"",code_qput,FATAL);
    (void) code_qput("\n  ======> ",11);
    (void) code_qput(strnum,fmt_ulong(strnum,i));
    (void) code_qput("\n",1);
    copybottom();
    qmail_to(&qq,pmod);

  } else if (act == AC_LOG) {
    if (!flaglist || !flagremote)
      strerr_die2x(100,FATAL,ERR_NOT_AVAILABLE);
    if (!pmod)
      strerr_die2x(100,FATAL,ERR_NOT_ALLOWED);
    if (!stralloc_copys(&line,workdir)) die_nomem();
    if (!stralloc_cats(&line,"/Log")) die_nomem();
    if (!stralloc_0(&line)) die_nomem();
    fd = open_read(line.s);
    if (fd == -1)
      if (errno != error_noent)
	strerr_die4sys(111,FATAL,ERR_OPEN,line.s,": ");
      else
        strerr_die3x(100,FATAL,line.s,ERR_NOEXIST);
    substdio_fdbuf(&sstext,read,fd,textbuf,sizeof(textbuf));
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    for (;;) {
      if (getln(&sstext,&line,&match,'\n') == -1)
        strerr_die2sys(111,FATAL,ERR_READ_INPUT);
      if (!match) break;
      (void) scan_ulong(line.s,&when);
      datetime_tai(&dt,when);		/* there is always at least a '\n' */
      (void) code_qput(date,date822fmt(date,&dt) - 1);
      (void) code_qput(": ",2);
      (void) code_qput(line.s,line.len);
    }
    close(fd);
    copybottom();
    qmail_to(&qq,pmod);

  } else if (act == AC_EDIT) {
	/* only remote admins and only if -e is specified may edit */
    if (!flagedit || !flagremote)
      strerr_die2x(100,FATAL,ERR_NOT_AVAILABLE);
    if (!pmod)
      strerr_die2x(100,FATAL,ERR_NOT_ALLOWED);
    len = str_len(ACTION_EDIT);
    if (!case_starts(action,ACTION_EDIT))
      len = str_len(ALT_EDIT);

    if (action[len]) {			/* -edit.file, not just -edit */
      if (action[len] != '.')
        strerr_die2x(100,FATAL,ERR_BAD_REQUEST);
      if (!stralloc_copys(&fnedit,"text/")) die_nomem();
      if (!stralloc_cats(&fnedit,action+len+1)) die_nomem();
      if (!stralloc_0(&fnedit)) die_nomem();
      case_lowerb(fnedit.s,fnedit.len);
      i = 5;	/* after the "text/" */
      while ((ch = fnedit.s[i++])) {
        if (((ch > 'z') || (ch < 'a')) && (ch != '_'))
          strerr_die2x(100,FATAL,ERR_BAD_NAME);
        if (ch == '_') fnedit.s[i-1] = '-';
      }
      switch(slurp(fnedit.s,&text,1024)) {	/* entire file! */
        case -1:
          strerr_die6sys(111,FATAL,ERR_READ,dir,"/",fnedit.s,": ");
        case 0:
          strerr_die5x(100,FATAL,dir,"/",fnedit.s,ERR_NOEXIST);
      }
      if (!stralloc_copy(&line,&text)) die_nomem();
      {		/* get rid of nulls to use cookie */
        register char *s; register unsigned int n;
        s = line.s; n = line.len;
        while(n--) { if (!*s) *s = '_'; ++s; }
      }
      if (!stralloc_cat(&line,&fnedit)) die_nomem();	/* including '\0' */
      strnum[fmt_ulong(strnum,(unsigned long) when)] = 0;
      cookie(hash,key.s,key.len,strnum,line.s,"-e");
      if (!stralloc_copy(&confirm,&outlocal)) die_nomem();
      if (!stralloc_catb(&confirm,ACTION_ED,LENGTH_ED)) die_nomem();
      if (!stralloc_cats(&confirm,strnum)) die_nomem();
      if (!stralloc_append(&confirm,".")) die_nomem();
		/* action part has been checked for bad chars */
      if (!stralloc_cats(&confirm,action + len + 1)) die_nomem();
      if (!stralloc_append(&confirm,".")) die_nomem();
      if (!stralloc_catb(&confirm,hash,COOKIE)) die_nomem();
      if (!stralloc_append(&confirm,"@")) die_nomem();
      if (!stralloc_cat(&confirm,&outhost)) die_nomem();
      if (!stralloc_0(&confirm)) die_nomem();
      set_cpconfirm(confirm.s);

      qmail_puts(&qq,"Reply-To: ");
      if (!quote2(&quoted,confirm.s)) die_nomem();
      qmail_put(&qq,quoted.s,quoted.len);
      qmail_puts(&qq,"\n");
      if (!stralloc_0(&confirm)) die_nomem();

      qmail_puts(&qq,TXT_EDIT_RESPONSE);
      qmail_puts(&qq,action+len+1);	/* has the '_' not '-' */
      qmail_puts(&qq,TXT_EDIT_FOR);
      if (!quote(&quoted,&outlocal)) die_nomem();
      qmail_put(&qq,quoted.s,quoted.len);
      qmail_puts(&qq,"@");
      qmail_put(&qq,outhost.s,outhost.len);
      qmail_puts(&qq,"\n");
      transferenc();
      copy(&qq,"text/top",flagcd,FATAL);
      copy(&qq,"text/edit-do",flagcd,FATAL);
      (void) code_qput(TXT_EDIT_START,str_len(TXT_EDIT_START));
      (void) code_qput("\n",1);
      (void) code_qput(text.s,text.len);
      (void) code_qput(TXT_EDIT_END,str_len(TXT_EDIT_END));
      (void) code_qput("\n",1);

    } else {	/* -edit only, so output list of editable files */
      qmail_puts(&qq,TXT_EDIT_LIST);
      transferenc();
      copy(&qq,"text/top",flagcd,FATAL);
      copy(&qq,"text/edit-list",flagcd,FATAL);
    }
    qmail_puts(&qq,"\n\n");
    copybottom();
    qmail_to(&qq,pmod);

  } else if (str_start(action,ACTION_ED)) {
    datetime_sec u;
    int flaggoodfield;
    x = action + LENGTH_ED;
    x += scan_ulong(x,&u);
    if ((u > when) || (u < when - 100000)) die_cookie();
    if (*x == '.') ++x;
    fname = x;
    x += str_chr(x,'.');
    if (!*x) die_cookie();
    *x = (char) 0;
    ++x;
    if (!stralloc_copys(&fnedit,"text/")) die_nomem();
    if (!stralloc_cats(&fnedit,fname)) die_nomem();
    if (!stralloc_0(&fnedit)) die_nomem();
    y = fnedit.s + 5;		/* after "text/" */
    while (*++y) {		/* Name should be guaranteed by the cookie, */
				/* but better safe than sorry ... */
      if (((*y > 'z') || (*y < 'a')) && (*y != '_'))
          strerr_die2x(100,FATAL,ERR_BAD_NAME);
      if (*y == '_') *y = '-';
    }

    fdlock = open_append("lock");
    if (fdlock == -1)
      strerr_die4sys(111,FATAL,ERR_OPEN,dir,"/lock: ");
    if (lock_ex(fdlock) == -1)
      strerr_die4sys(111,FATAL,ERR_OBTAIN,dir,"/lock: ");

    switch (slurp(fnedit.s,&text,1024)) {
      case -1:
        strerr_die6sys(111,FATAL,ERR_READ,dir,"/",fnedit.s,": ");
      case 0:
        strerr_die5x(100,FATAL,dir,"/",fnedit.s,ERR_NOEXIST);
    }
    if (!stralloc_copy(&line,&text)) die_nomem();
    {		/* get rid of nulls to use cookie */
      register char *s; register unsigned int n;
      s = line.s; n = line.len;
      while(n--) { if (!*s) *s = '_'; ++s; }
    }
    if (!stralloc_cat(&line,&fnedit)) die_nomem();	/* including '\0' */
    strnum[fmt_ulong(strnum,(unsigned long) u)] = 0;
    cookie(hash,key.s,key.len,strnum,line.s,"-e");
    if (str_len(x) != COOKIE) die_cookie();
    if (byte_diff(hash,COOKIE,x)) die_cookie();
	/* cookie is ok, file exists, lock's on, new file ends in '_' */
    if (!stralloc_copys(&fneditn,fnedit.s)) die_nomem();
    if (!stralloc_append(&fneditn,"_")) die_nomem();
    if (!stralloc_0(&fneditn)) die_nomem();
    fd = open_trunc(fneditn.s);
    if (fd == -1)
      strerr_die6sys(111,FATAL,ERR_WRITE,dir,"/",fneditn.s,": ");
    substdio_fdbuf(&sstext,write,fd,textbuf,sizeof(textbuf));
    if (!stralloc_copys(&quoted,"")) die_nomem();	/* clear */
    if (!stralloc_copys(&text,"")) die_nomem();

    for (;;) {			/* get message body */
      if (getln(&ssin,&line,&match,'\n') == -1)
        strerr_die2sys(111,FATAL,ERR_READ_INPUT);
      if (!match) break;
      if (!stralloc_cat(&text,&line)) die_nomem();
    }
    if (encin) {	/* decode if necessary */
      if (encin == 'B')
        decodeB(text.s,text.len,&line,FATAL);
      else
        decodeQ(text.s,text.len,&line,FATAL);
      if (!stralloc_copy(&text,&line)) die_nomem();
    }
    cp = text.s;
    cpafter = text.s+text.len;
    flaggoodfield = 0;
    flagdone = 0;
    len = 0;
    while ((cpnext = cp + byte_chr(cp,cpafter-cp,'\n')) != cpafter) {
      i = byte_chr(cp,cpnext-cp,'%');
      if (i != (unsigned int) (cpnext - cp)) {
        if (!flaggoodfield) {	/* TXT_EDIT_START/END */
          if (case_startb(cp+i,cpnext-cp-i,TXT_EDIT_START)) {
		/* start tag. Store users 'quote characters', e.g. '> ' */
            if (!stralloc_copyb(&quoted,cp,i)) die_nomem();
            flaggoodfield = 1;
            cp = cpnext + 1;
            cpfirst = cp;
            continue;
          }
        } else
          if (case_startb(cp+i,cpnext-cp-i,TXT_EDIT_END)) {
            flagdone = 1;
            break;
          }
      }
      if (flaggoodfield) {
        if ((len += cpnext - cp - quoted.len + 1) > MAXEDIT)
          strerr_die1x(100,ERR_EDSIZE);

        if (quoted.len && cpnext-cp >= (int) quoted.len &&
			!str_diffn(cp,quoted.s,quoted.len))
          cp += quoted.len;	/* skip quoting characters */
        cplast = cpnext - 1;
        if (*cplast == '\r')	/* CRLF -> '\n' for base64 encoding */
          *cplast = '\n';
        else
          ++cplast;
        if (substdio_put(&sstext,cp,cplast-cp+1) == -1)
            strerr_die6sys(111,FATAL,ERR_WRITE,dir,"/",fneditn.s,": ");
      }
      cp = cpnext + 1;
    }
    if (!flagdone)
      strerr_die2x(100,FATAL,ERR_NO_MARK);
    if (substdio_flush(&sstext) == -1)
      strerr_die6sys(111,FATAL,ERR_WRITE,dir,"/",fneditn.s,": ");
    if (fsync(fd) == -1)
      strerr_die6sys(111,FATAL,ERR_SYNC,dir,"/",fneditn.s,": ");
    if (fchmod(fd, 0600) == -1)
      strerr_die6sys(111,FATAL,ERR_CHMOD,dir,"/",fneditn.s,": ");
    if (close(fd) == -1)
      strerr_die6sys(111,FATAL,ERR_CLOSE,dir,"/",fneditn.s,": ");
    if (rename(fneditn.s,fnedit.s) == -1)
      strerr_die6sys(111,FATAL,ERR_MOVE,dir,"/",fneditn.s,": ");

    close(fdlock);
    qmail_puts(&qq,TXT_EDIT_SUCCESS);
    qmail_puts(&qq,fname);
    qmail_puts(&qq,TXT_EDIT_FOR);
    if (!quote(&quoted,&outlocal)) die_nomem();
    qmail_put(&qq,quoted.s,quoted.len);
    qmail_puts(&qq,"@");
    qmail_put(&qq,outhost.s,outhost.len);
    qmail_puts(&qq,"\n");
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    copy(&qq,"text/edit-done",flagcd,FATAL);
    copybottom();
    qmail_to(&qq,sender);	/* not necessarily from mod */

  } else if (act == AC_GET) {

    unsigned long u;
    struct stat st;
    char ch;
    int r;
    unsigned int pos;

    if (!flagget)
      strerr_die2x(100,FATAL,ERR_NOT_AVAILABLE);
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);

    pos = str_len(ACTION_GET);
    if (!case_starts(action,ACTION_GET))
      pos = str_len(ALT_GET);

    if (action[pos] == '.' || action [pos] == '_') pos++;
    scan_ulong(action + pos,&u);

    if (!stralloc_copys(&line,"archive/")) die_nomem();
    if (!stralloc_catb(&line,strnum,fmt_ulong(strnum,u / 100))) die_nomem();
    if (!stralloc_cats(&line,"/")) die_nomem();
    if (!stralloc_catb(&line,strnum,fmt_uint0(strnum,(unsigned int) (u % 100),2))) die_nomem();
    if (!stralloc_0(&line)) die_nomem();

    fd = open_read(line.s);
    if (fd == -1)
      if (errno != error_noent)
	strerr_die4sys(111,FATAL,ERR_OPEN,line.s,": ");
      else
        copy(&qq,"text/get-bad",flagcd,FATAL);
    else {
      if (fstat(fd,&st) == -1)
        copy(&qq,"text/get-bad",flagcd,FATAL);
      else if (!(st.st_mode & 0100))
        copy(&qq,"text/get-bad",flagcd,FATAL);
      else {
        substdio_fdbuf(&sstext,read,fd,textbuf,sizeof(textbuf));
	qmail_puts(&qq,"> ");
	for (;;) {
	  r = substdio_get(&sstext,&ch,1);
	  if (r == -1) strerr_die4sys(111,FATAL,ERR_READ,line.s,": ");
	  if (r == 0) break;
	  qmail_put(&qq,&ch,1);
	  if (ch == '\n') qmail_puts(&qq,"> ");
	}
	qmail_puts(&qq,"\n");
      }
      close(fd);
    }
    copybottom();
    qmail_to(&qq,target.s);

  } else if (case_starts(action,ACTION_QUERY) ||
		case_starts(action,ALT_QUERY)) {
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    if (pmod) {
      if (!stralloc_copys(&to,pmod)) die_nomem();
      if (!stralloc_0(&to)) die_nomem();
    } else
      if (!stralloc_copy(&to,&target)) die_nomem();
    if (issub(workdir,target.s,FATAL))
      copy(&qq,"text/sub-nop",flagcd,FATAL);
    else
      copy(&qq,"text/unsub-nop",flagcd,FATAL);
    copybottom();
    qmail_to(&qq,to.s);

  } else if (case_starts(action,ACTION_INFO) ||
		case_starts(action,ALT_INFO)) {
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    copy(&qq,"text/info",flagcd,FATAL);
    copybottom();
    qmail_to(&qq,target.s);

  } else if (case_starts(action,ACTION_FAQ) ||
		case_starts(action,ALT_FAQ)) {
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    copy(&qq,"text/faq",flagcd,FATAL);
    copybottom();
    qmail_to(&qq,target.s);

  } else if (pmod && (act == AC_HELP)) {
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    copy(&qq,"text/mod-help",flagcd,FATAL);
    copybottom();
    qmail_to(&qq,pmod);

  } else {
    qmail_puts(&qq,TXT_EZMLM_RESPONSE);
    transferenc();
    copy(&qq,"text/top",flagcd,FATAL);
    copy(&qq,"text/help",flagcd,FATAL);
    copybottom();
    qmail_to(&qq,sender);
  }

  switch(qmail_close(&qq)) {
    case 0:
      strnum[fmt_ulong(strnum,qmail_qp(&qq))] = 0;
      strerr_die2x(0,"ezmlm-manage: info: qp ",strnum);
    default:
      strerr_die2x(111,FATAL,ERR_TMP_QMAIL_QUEUE);
  }
}

